""" $lic$
Copyright (c) 2016-2017, Mingyu Gao
All rights reserved.

This program is free software: you can redistribute it and/or modify it under
the terms of the Modified BSD-3 License as published by the Open Source
Initiative.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the BSD-3 License for more details.

You should have received a copy of the Modified BSD-3 License along with this
program. If not, see <https://opensource.org/licenses/BSD-3-Clause>.
"""

from .timing import Timing
from .voltage_domain import VoltageDomain


class EnergyLPDDR(object):
    '''
    Calculate energy for LPDDR2, LPDDR3.
    '''

    def __init__(self, tck, timing, vdd1, idds1, vdd2, idds2, vddcaq, iddsin,
                 chipcnt, ddr=3):
        if ddr == 2 or ddr == 3:
            if round(vdd1, 5) != round(1.8, 5):
                raise ValueError('{}: given vdd1 is invalid.'
                                 .format(self.__class__.__name__)
                                 + ' should be 1.8 V for LPDDR2/3.')
            if round(vdd2, 5) != round(1.2, 5):
                raise ValueError('{}: given vdd2 is invalid.'
                                 .format(self.__class__.__name__)
                                 + ' should be 1.2 V for LPDDR2/3.')
            if round(vddcaq, 5) != round(1.2, 5):
                raise ValueError('{}: given vddcaq is invalid.'
                                 .format(self.__class__.__name__)
                                 + ' should be 1.2 V for LPDDR2/3.')
            self.type = 'LPDDR{}'.format(ddr)
        elif ddr >= 4:
            raise ValueError('{}: given ddr is invalid.'
                             .format(self.__class__.__name__)
                             + ' LPDDR4 and above are not supported currently.')
        else:
            raise ValueError('{}: given ddr is invalid.'
                             .format(self.__class__.__name__))

        if not isinstance(timing, Timing):
            raise TypeError('{}: given timing has invalid type.'
                            .format(self.__class__.__name__))
        self.timing = timing
        self.vdom1 = VoltageDomain(tck, vdd1, idds1, chipcnt, ddr=ddr)
        self.vdom2 = VoltageDomain(tck, vdd2, idds2, chipcnt, ddr=ddr)
        self.vdomcaq = VoltageDomain(tck, vddcaq, iddsin, chipcnt, ddr=ddr)

    def background_energy(self, cycles_bankpre_ckelo=0, cycles_bankpre_ckehi=0,
                          cycles_bankact_ckelo=0, cycles_bankact_ckehi=0):
        ''' Background energy. '''
        return self.vdom1.background_energy(
            cycles_bankpre_ckelo=cycles_bankpre_ckelo,
            cycles_bankpre_ckehi=cycles_bankpre_ckehi,
            cycles_bankact_ckelo=cycles_bankact_ckelo,
            cycles_bankact_ckehi=cycles_bankact_ckehi) \
        + self.vdom2.background_energy(
            cycles_bankpre_ckelo=cycles_bankpre_ckelo,
            cycles_bankpre_ckehi=cycles_bankpre_ckehi,
            cycles_bankact_ckelo=cycles_bankact_ckelo,
            cycles_bankact_ckehi=cycles_bankact_ckehi) \
        + self.vdomcaq.background_energy(
            cycles_bankpre_ckelo=cycles_bankpre_ckelo,
            cycles_bankpre_ckehi=cycles_bankpre_ckehi,
            cycles_bankact_ckelo=cycles_bankact_ckelo,
            cycles_bankact_ckehi=cycles_bankact_ckehi)

    def activate_energy(self, num_act=1):
        ''' Activate energy. '''
        return self.vdom1.activate_energy(self.timing, num_act=num_act) \
                + self.vdom2.activate_energy(self.timing, num_act=num_act) \
                + self.vdomcaq.activate_energy(self.timing, num_act=num_act)

    def readwrite_energy(self, num_rd=1, num_wr=0):
        ''' Read write energy. '''
        return self.vdom1.readwrite_energy(num_rd=num_rd, num_wr=num_wr) \
                + self.vdom2.readwrite_energy(num_rd=num_rd, num_wr=num_wr) \
                + self.vdomcaq.readwrite_energy(num_rd=num_rd, num_wr=num_wr)

    def refresh_energy(self, num_ref=1):
        ''' Refresh energy. '''
        return self.vdom1.refresh_energy(self.timing, num_ref=num_ref) \
                + self.vdom2.refresh_energy(self.timing, num_ref=num_ref) \
                + self.vdomcaq.refresh_energy(self.timing, num_ref=num_ref)




""" $lic$
Copyright (c) 2016-2017, Mingyu Gao
All rights reserved.

This program is free software: you can redistribute it and/or modify it under
the terms of the Modified BSD-3 License as published by the Open Source
Initiative.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the BSD-3 License for more details.

You should have received a copy of the Modified BSD-3 License along with this
program. If not, see <https://opensource.org/licenses/BSD-3-Clause>.
"""

from .energy_ddr import EnergyDDR
from .energy_lpddr import EnergyLPDDR
from .termination import TermResistance, Termination
from .timing import Timing
from .voltage_domain import IDDs, VoltageDomain

__version__ = '0.3.0'




""" $lic$
Copyright (c) 2016-2017, Mingyu Gao
All rights reserved.

This program is free software: you can redistribute it and/or modify it under
the terms of the Modified BSD-3 License as published by the Open Source
Initiative.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the BSD-3 License for more details.

You should have received a copy of the Modified BSD-3 License along with this
program. If not, see <https://opensource.org/licenses/BSD-3-Clause>.
"""

from collections import namedtuple

_IDDsBase = namedtuple('_IDDsBase', ['idd0', 'idd2n', 'idd2p', 'idd3n',
                                     'idd3p', 'idd4r', 'idd4w', 'idd5'])

class IDDs(_IDDsBase):
    '''
    Define the set of IDD values.
    '''

    def __new__(cls, **kwargs):
        # http://stackoverflow.com/a/3624799
        self = super(IDDs, cls).__new__(cls, **kwargs)
        self.check()
        return self

    def check(self):
        ''' Check validation of IDD values. '''
        if self.idd2n < self.idd2p:
            raise ValueError('{}: IDD2N < IDD2P!'
                             .format(self.__class__.__name__))
        if self.idd3n < self.idd3p:
            raise ValueError('{}: IDD3N < IDD3P!'
                             .format(self.__class__.__name__))
        if self.idd3n < self.idd2n:
            raise ValueError('{}: IDD3N < IDD2N!'
                             .format(self.__class__.__name__))
        if self.idd3p < self.idd2p:
            raise ValueError('{}: IDD3P < IDD2P!'
                             .format(self.__class__.__name__))
        if self.idd0 < self.idd3n:
            raise ValueError('{}: IDD0 < IDD3N!'
                             .format(self.__class__.__name__))
        if self.idd4r < self.idd3n:
            raise ValueError('{}: IDD4R < IDD3N!'
                             .format(self.__class__.__name__))
        if self.idd4w < self.idd3n:
            raise ValueError('{}: IDD4W < IDD3N!'
                             .format(self.__class__.__name__))
        if self.idd5 < self.idd3n:
            raise ValueError('{}: IDD5 < IDD3N!'
                             .format(self.__class__.__name__))


class VoltageDomain(object):
    '''
    Define a voltage domain including VDD and IDD values.
    '''

    def __init__(self, tck, vdd, idds, chipcnt, ddr=3):
        if tck < 0:
            raise ValueError('{}: given tck is invalid.'
                             .format(self.__class__.__name__))
        if vdd < 0:
            raise ValueError('{}: given vdd is invalid.'
                             .format(self.__class__.__name__))
        if not isinstance(idds, IDDs):
            raise TypeError('{}: given idds has invalid type.'
                            .format(self.__class__.__name__))
        if not isinstance(chipcnt, int):
            raise TypeError('{}: given chipcnt has invalid type.'
                            .format(self.__class__.__name__))
        if chipcnt <= 0:
            raise ValueError('{}: given chipcnt is invalid.'
                             .format(self.__class__.__name__))
        if not isinstance(ddr, int):
            raise TypeError('{}: given ddr has invalid type.'
                            .format(self.__class__.__name__))
        if ddr < 0:
            raise ValueError('{}: given ddr is invalid.'
                             .format(self.__class__.__name__))
        self.tck = tck
        self.vdd = vdd
        self.idds = idds
        self.chipcnt = chipcnt
        self.ddr = ddr

    def background_energy(self, cycles_bankpre_ckelo=0, cycles_bankpre_ckehi=0,
                          cycles_bankact_ckelo=0, cycles_bankact_ckehi=0):
        ''' Background energy. '''
        chipicyc = (self.idds.idd2p * cycles_bankpre_ckelo
                    + self.idds.idd2n * cycles_bankpre_ckehi
                    + self.idds.idd3p * cycles_bankact_ckelo
                    + self.idds.idd3n * cycles_bankact_ckehi)
        return chipicyc * self.vdd * self.tck * self.chipcnt

    def activate_energy(self, timing, num_act=1):
        ''' Activate energy. '''
        cycles_rc = (timing.RAS + timing.RP)
        chipicyc = num_act * (self.idds.idd0 * cycles_rc
                              - self.idds.idd3n * timing.RAS
                              - self.idds.idd2n * (cycles_rc - timing.RAS))
        return chipicyc * self.vdd * self.tck * self.chipcnt

    def readwrite_energy(self, num_rd=1, num_wr=0):
        ''' Read write energy. '''
        cyc_burst = (1 << self.ddr) / 2 if self.ddr >= 1 else 1
        chipicyc = cyc_burst * ((self.idds.idd4r - self.idds.idd3n) * num_rd
                                + (self.idds.idd4w - self.idds.idd3n) * num_wr)
        return chipicyc * self.vdd * self.tck * self.chipcnt

    def refresh_energy(self, timing, num_ref=1):
        ''' Refresh energy. '''
        chipicyc = timing.RFC * (self.idds.idd5 - self.idds.idd3n) * num_ref
        return chipicyc * self.vdd * self.tck * self.chipcnt




""" $lic$
Copyright (c) 2016-2017, Mingyu Gao
All rights reserved.

This program is free software: you can redistribute it and/or modify it under
the terms of the Modified BSD-3 License as published by the Open Source
Initiative.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the BSD-3 License for more details.

You should have received a copy of the Modified BSD-3 License along with this
program. If not, see <https://opensource.org/licenses/BSD-3-Clause>.
"""

from collections import namedtuple

# Alphabetical order.
_TIMING_PARAM_LIST = [
    'RAS',
    'REFI',
    'RFC',
    'RP',
    'RRD',
    ]

'''
Define timing parameters in unit of cycles.
'''
Timing = namedtuple('Timing', _TIMING_PARAM_LIST)




""" $lic$
Copyright (c) 2016-2017, Mingyu Gao
All rights reserved.

This program is free software: you can redistribute it and/or modify it under
the terms of the Modified BSD-3 License as published by the Open Source
Initiative.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the BSD-3 License for more details.

You should have received a copy of the Modified BSD-3 License along with this
program. If not, see <https://opensource.org/licenses/BSD-3-Clause>.
"""

from collections import namedtuple
import numpy as np

_TermResistanceBase = namedtuple('_TermResistanceBase',
                                 ['rz_dev', 'rz_mc', 'rtt_nom', 'rtt_wr',
                                  'rtt_mc', 'rs'])

class TermResistance(_TermResistanceBase):
    '''
    Define the set of termination resistance values.

    rz_dev: DRAM device output driver impedance.
    rz_mc: memory controller output driver impedance.
    rtt_nom: nominal termination.
    rtt_wr: dynamic write termination.
    rtt_mc: memory controller termination.
    rs: trace impedance.
    '''

    def __new__(cls, **kwargs):
        # http://stackoverflow.com/a/3624799
        self = super(TermResistance, cls).__new__(cls, **kwargs)
        self.check()
        return self

    def check(self):
        ''' Check validation of termination resistance values. '''
        if self.rz_dev < 1e-4:
            raise ValueError('{}: given device driver impedance rz_dev '
                             'is invalid.'
                             .format(self.__class__.__name__))
        if self.rz_mc < 1e-4:
            raise ValueError('{}: given memory ctlr driver impedance rz_mc '
                             'is invalid.'
                             .format(self.__class__.__name__))
        if self.rtt_nom < 1e-4:
            raise ValueError('{}: given device R_TT,nom is invalid.'
                             .format(self.__class__.__name__))
        if self.rtt_wr < 1e-4:
            raise ValueError('{}: given device R_TT(WR) is invalid.'
                             .format(self.__class__.__name__))
        if self.rtt_mc < 1e-4:
            raise ValueError('{}: given memory ctlr termination rtt_mc '
                             'is invalid.'
                             .format(self.__class__.__name__))
        if self.rs < 1e-4:
            raise ValueError('{}: given trace impedance rs is invalid.'
                             .format(self.__class__.__name__))


class Termination(object):
    '''
    Termination scheme for an individual chip.
    '''

    def __init__(self, vdd, rankcnt, resistance, width=0, level='mid',
                 with_dqs=True, with_dm=True, with_dbi=False):
        '''
        `width` specifies the chip width and determines the pin count
        associated to termination. Currently support 0, 4, 8, 16, 32. Valid for
        DDR2, DDR3, DDR4, LPDDR3, and GDDR5.

        Should be configured with `with_dqs`, `with_dm`, `with_dbi` to specify
        the presence of DQS, DM, DBI pins.

        DDR2/3/4 and LPDDR3 have DQS and DM but no DBI.

        GDDR5 does not have DQS or DM, but has DBI.

        A special value 0 for `width` means to calculate for a single pin.

        `level` can be 'high' or 'mid'. DDR2 and DDR3 use mid; DDR4 uses high.

        high: single R_TT connects to VDD.

        mid: R_TTU connects to VDD and R_TTD connects to GND, both are 2 * R_TT.
        '''

        if vdd < 0:
            raise ValueError('{}: given vdd is invalid.'
                             .format(self.__class__.__name__))
        if not isinstance(rankcnt, int):
            raise TypeError('{}: given rankcnt has invalid type.'
                            .format(self.__class__.__name__))
        if rankcnt <= 0:
            raise ValueError('{}: given rankcnt is invalid.'
                             .format(self.__class__.__name__))
        if not isinstance(resistance, TermResistance):
            raise TypeError('{}: given resistance has invalid type.'
                            .format(self.__class__.__name__))

        self.vdd = vdd
        self.rankcnt = rankcnt
        self.resistance = resistance

        if width == 0:
            self.rdpincnt = 1
            self.wrpincnt = 1
        elif width >= 4 and ((width & (width - 1)) == 0):
            # Width must be power of 2.

            self.rdpincnt = 0
            self.wrpincnt = 0
            # DQ switch is halved, DBI for each eight-pin group.
            if with_dbi:
                self.rdpincnt += width / 2 + max(1, width / 8)
                self.wrpincnt += width / 2 + max(1, width / 8)
            else:
                self.rdpincnt += width
                self.wrpincnt += width
            # DQS, DQS# for each eight-pin group.
            self.rdpincnt += (max(1, width / 8) * 2 if with_dqs else 0)
            self.wrpincnt += (max(1, width / 8) * 2 if with_dqs else 0)
            # DM for each eight-pin group.
            self.wrpincnt += (max(1, width / 8) if with_dm else 0)
        else:
            raise ValueError('{}: given width is invalid.'
                             .format(self.__class__.__name__))

        if level == 'high':
            vdd_eq = vdd / 1.
        elif level == 'mid':
            vdd_eq = vdd / 2.
        else:
            raise ValueError('{}: given level is invalid.'
                             .format(self.__class__.__name__))

        rz_dev = resistance.rz_dev
        rz_mc = resistance.rz_mc
        rtt_nom = resistance.rtt_nom
        rtt_wr = resistance.rtt_wr
        rtt_mc = resistance.rtt_mc
        rs = resistance.rs

        # According to nodal-voltage analysis, solve coef * vnodes = rhs, where
        # vnodes = [V0, V1, V2, ..., VMC], for DRAM read and DRAM write,
        # respectively.

        # DRAM read.

        coef = np.zeros((rankcnt + 1, rankcnt + 1))
        rhs = np.zeros(rankcnt + 1)
        # Rank 0: rz_dev, to GND.
        # V0/rz_dev + (V0 - VMC)/rs = 0
        coef[0, 0] = 1. / rz_dev + 1. / rs
        coef[0, rankcnt] = -1. / rs
        rhs[0] = 0
        # Other rankcnt: rtt_nom, 2x up to VDD and down to GND.
        # (Vi - VDD)/2rtt_nom + Vi/2rtt_nom + (Vi - VMC)/rs = 0
        for idx in range(1, rankcnt):
            coef[idx, idx] = 1. / rtt_nom + 1. / rs
            coef[idx, rankcnt] = -1. / rs
            rhs[idx] = vdd_eq / rtt_nom
        # MC: rtt_mc, 2x up to VDD and down to GND.
        # (VMC - VDD)/2rtt_mc + VMC/2rtt_mc + sum (VMC - Vi)/rs = 0
        coef[rankcnt, 0:rankcnt] = -1. / rs
        coef[rankcnt, rankcnt] = 1. / rtt_mc + rankcnt * 1. / rs
        rhs[rankcnt] = vdd_eq / rtt_mc

        rd_vnodes = np.linalg.solve(coef, rhs)

        self.rd_power = np.zeros(rankcnt + 1)
        self.rd_power[0] = \
                (rd_vnodes[0] ** 2) / rz_dev \
                + ((rd_vnodes[0] - rd_vnodes[-1]) ** 2) / rs
        for idx in range(1, rankcnt):
            if level == 'mid':
                self.rd_power[idx] = \
                        ((rd_vnodes[idx] - vdd) ** 2) / 2 / rtt_nom \
                        + (rd_vnodes[idx] ** 2) / 2 / rtt_nom \
                        + ((rd_vnodes[idx] - rd_vnodes[-1]) ** 2) / rs
            elif level == 'high':
                self.rd_power[idx] = \
                        ((rd_vnodes[idx] - vdd) ** 2) / rtt_nom \
                        + ((rd_vnodes[idx] - rd_vnodes[-1]) ** 2) / rs
        if level == 'mid':
            self.rd_power[-1] = \
                    ((rd_vnodes[-1] - vdd) ** 2) / 2 / rtt_mc \
                    + (rd_vnodes[-1] ** 2) / 2 / rtt_mc
        elif level == 'high':
            self.rd_power[-1] = ((rd_vnodes[-1] - vdd) ** 2) / rtt_mc

        # DRAM write.

        coef = np.zeros((rankcnt + 1, rankcnt + 1))
        rhs = np.zeros(rankcnt + 1)
        # Rank 0: rtt_wr, 2x up to VDD and down to GND.
        # (V0 - VDD)/2rtt_wr + V0/2rtt_wr + (V0 - VMC)/rs = 0
        coef[0, 0] = 1. / rtt_wr + 1. / rs
        coef[0, rankcnt] = -1. / rs
        rhs[0] = vdd_eq / rtt_wr
        # Other rankcnt: rtt_nom, 2x up to VDD and down to GND.
        # (Vi - VDD)/2rtt_nom + Vi/2rtt_nom + (Vi - VMC)/rs = 0
        for idx in range(1, rankcnt):
            coef[idx, idx] = 1. / rtt_nom + 1. / rs
            coef[idx, rankcnt] = -1. / rs
            rhs[idx] = vdd_eq / rtt_nom
        # MC: rz_mc, to GND.
        # VMC/rz_mc + sum (VMC - Vi)/rs = 0
        coef[rankcnt, 0:rankcnt] = -1. / rs
        coef[rankcnt, rankcnt] = 1. / rz_mc + rankcnt * 1. / rs
        rhs[rankcnt] = 0

        wr_vnodes = np.linalg.solve(coef, rhs)

        self.wr_power = np.zeros(rankcnt + 1)
        if level == 'mid':
            self.wr_power[0] = \
                    ((wr_vnodes[0] - vdd) ** 2) / 2 / rtt_wr \
                    + (wr_vnodes[0] ** 2) / 2 / rtt_wr \
                    + ((wr_vnodes[0] - wr_vnodes[-1]) ** 2) / rs
        elif level == 'high':
            self.wr_power[0] = \
                    ((wr_vnodes[0] - vdd) ** 2) / rtt_wr \
                    + ((wr_vnodes[0] - wr_vnodes[-1]) ** 2) / rs
        for idx in range(1, rankcnt):
            if level == 'mid':
                self.wr_power[idx] = \
                        ((wr_vnodes[idx] - vdd) ** 2) / 2 / rtt_nom \
                        + (wr_vnodes[idx] ** 2) / 2 / rtt_nom \
                        + ((wr_vnodes[idx] - wr_vnodes[-1]) ** 2) / rs
            elif level == 'high':
                self.wr_power[idx] = \
                        ((wr_vnodes[idx] - vdd) ** 2) / rtt_nom \
                        + ((wr_vnodes[idx] - wr_vnodes[-1]) ** 2) / rs
        self.wr_power[-1] = (wr_vnodes[-1] ** 2) / rz_mc

        # Multiply pin count to be a whole chip.
        self.rd_power *= self.rdpincnt
        self.wr_power *= self.wrpincnt

    def read_power_total(self):
        ''' Get DRAM read termination power. '''
        return self.rd_power.sum()

    def write_power_total(self):
        ''' Get DRAM write termination power. '''
        return self.wr_power.sum()

    def read_power_memctlr(self):
        ''' Get DRAM read termination power at memory controller. '''
        return self.rd_power[-1]

    def write_power_memctlr(self):
        ''' Get DRAM write termination power at memory controller. '''
        return self.wr_power[-1]

    def read_power_devices(self):
        ''' Get DRAM read termination power at DRAM devices. '''
        return self.rd_power[:self.rankcnt].sum()

    def write_power_devices(self):
        ''' Get DRAM write termination power at DRAM devices. '''
        return self.wr_power[:self.rankcnt].sum()

    def read_power_target_rank(self):
        ''' Get DRAM read termination power at the target rank. '''
        return self.rd_power[0]

    def write_power_target_rank(self):
        ''' Get DRAM write termination power at the target rank. '''
        return self.wr_power[0]

    def read_power_other_ranks(self):
        ''' Get DRAM read termination power at other ranks. '''
        return self.rd_power[1:self.rankcnt].sum() if self.rankcnt > 1 else 0

    def write_power_other_ranks(self):
        ''' Get DRAM write termination power at other ranks. '''
        return self.wr_power[1:self.rankcnt].sum() if self.rankcnt > 1 else 0




""" $lic$
Copyright (c) 2016-2017, Mingyu Gao
All rights reserved.

This program is free software: you can redistribute it and/or modify it under
the terms of the Modified BSD-3 License as published by the Open Source
Initiative.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the BSD-3 License for more details.

You should have received a copy of the Modified BSD-3 License along with this
program. If not, see <https://opensource.org/licenses/BSD-3-Clause>.
"""

import unittest

import energydram


class TestIDDs(unittest.TestCase):
    ''' IDDs class unit tests. '''
    def test_init(self):
        ''' Initialization. '''
        idds = energydram.IDDs(idd0=115, idd2p=25, idd2n=65, idd3p=45,
                               idd3n=75, idd4r=220, idd4w=240, idd5=255)
        self.assertEqual(idds.idd0, 115, 'IDD0')
        self.assertEqual(idds.idd2p, 25, 'IDD2P')
        self.assertEqual(idds.idd2n, 65, 'IDD2N')
        self.assertEqual(idds.idd3p, 45, 'IDD3P')
        self.assertEqual(idds.idd3n, 75, 'IDD3N')
        self.assertEqual(idds.idd4r, 220, 'IDD4R')
        self.assertEqual(idds.idd4w, 240, 'IDD4W')
        self.assertEqual(idds.idd5, 255, 'IDD5')

    def test_init_2n_vs_2p(self):
        ''' Assert IDD2N >= IDD2P. '''
        with self.assertRaisesRegexp(ValueError, 'IDDs: .*2N.*2P.*'):
            energydram.IDDs(idd0=115, idd2p=25, idd2n=15, idd3p=45,
                            idd3n=75, idd4r=220, idd4w=240, idd5=255)

    def test_init_3n_vs_3p(self):
        ''' Assert IDD3N >= IDD3P. '''
        with self.assertRaisesRegexp(ValueError, 'IDDs: .*3N.*3P.*'):
            energydram.IDDs(idd0=115, idd2p=25, idd2n=65, idd3p=95,
                            idd3n=75, idd4r=220, idd4w=240, idd5=255)

    def test_init_3n_vs_2n(self):
        ''' Assert IDD3N >= IDD2N. '''
        with self.assertRaisesRegexp(ValueError, 'IDDs: .*3N.*2N.*'):
            energydram.IDDs(idd0=115, idd2p=25, idd2n=65, idd3p=45,
                            idd3n=55, idd4r=220, idd4w=240, idd5=255)

    def test_init_3p_vs_2p(self):
        ''' Assert IDD3P >= IDD2P. '''
        with self.assertRaisesRegexp(ValueError, 'IDDs: .*3P.*2P.*'):
            energydram.IDDs(idd0=115, idd2p=25, idd2n=65, idd3p=15,
                            idd3n=75, idd4r=220, idd4w=240, idd5=255)

    def test_init_0_vs_3n(self):
        ''' Assert IDD0 >= IDD3N. '''
        with self.assertRaisesRegexp(ValueError, 'IDDs: .*0.*3N.*'):
            energydram.IDDs(idd0=15, idd2p=25, idd2n=65, idd3p=45,
                            idd3n=75, idd4r=220, idd4w=240, idd5=255)

    def test_init_4r_vs_3n(self):
        ''' Assert IDD4R >= IDD3N. '''
        with self.assertRaisesRegexp(ValueError, 'IDDs: .*4R.*3N.*'):
            energydram.IDDs(idd0=115, idd2p=25, idd2n=65, idd3p=45,
                            idd3n=75, idd4r=70, idd4w=240, idd5=255)

    def test_init_4w_vs_3n(self):
        ''' Assert IDD4W >= IDD3N. '''
        with self.assertRaisesRegexp(ValueError, 'IDDs: .*4W.*3N.*'):
            energydram.IDDs(idd0=115, idd2p=25, idd2n=65, idd3p=45,
                            idd3n=75, idd4r=220, idd4w=70, idd5=255)

    def test_init_5_vs_3n(self):
        ''' Assert IDD5 >= IDD3N. '''
        with self.assertRaisesRegexp(ValueError, 'IDDs: .*5.*3N.*'):
            energydram.IDDs(idd0=115, idd2p=25, idd2n=65, idd3p=45,
                            idd3n=75, idd4r=220, idd4w=240, idd5=70)


class TestVoltageDomain(unittest.TestCase):
    '''
    VoltageDomain class unit tests.

    Compare with DDR3_Power_Calc.xlsm.

    Based on DDR3, 2 Gb, x8, -125E, fast-exit.
    '''

    tck = 1000./800
    vdd = 1.575
    idds = energydram.IDDs(idd0=95, idd2p=35, idd2n=42, idd3p=40,
                           idd3n=45, idd4r=180, idd4w=185, idd5=215)
    chipcnt = 1
    timing = energydram.Timing(RRD=6, RAS=35, RP=47.5-35, RFC=160, REFI=7800)

    def test_init(self):
        ''' Initialization. '''
        vdom = energydram.VoltageDomain(self.tck, self.vdd, self.idds,
                                        self.chipcnt)
        self.assertEqual(vdom.tck, self.tck, 'tck')
        self.assertEqual(vdom.vdd, self.vdd, 'vdd')
        self.assertEqual(vdom.idds, self.idds, 'idds')
        self.assertEqual(vdom.chipcnt, self.chipcnt, 'chipcnt')
        self.assertEqual(vdom.ddr, 3, 'ddr')

    def test_init_invalid_tck(self):
        ''' Initialize with invalid tck. '''
        with self.assertRaisesRegexp(ValueError, 'VoltageDomain: .*tck.*'):
            energydram.VoltageDomain(-1, self.vdd, self.idds, self.chipcnt)

    def test_init_invalid_vdd(self):
        ''' Initialize with invalid vdd. '''
        with self.assertRaisesRegexp(ValueError, 'VoltageDomain: .*vdd.*'):
            energydram.VoltageDomain(self.tck, -1.2, self.idds, self.chipcnt)

    def test_init_invalid_idds(self):
        ''' Initialize with invalid idds. '''
        with self.assertRaisesRegexp(TypeError, 'VoltageDomain: .*idds.*'):
            energydram.VoltageDomain(self.tck, self.vdd, None, self.chipcnt)

    def test_init_invalid_chipcnt(self):
        ''' Initialize with invalid chipcnt. '''
        with self.assertRaisesRegexp(TypeError, 'VoltageDomain: .*chipcnt.*'):
            energydram.VoltageDomain(self.tck, self.vdd, self.idds, None)

        with self.assertRaisesRegexp(ValueError, 'VoltageDomain: .*chipcnt.*'):
            energydram.VoltageDomain(self.tck, self.vdd, self.idds, 0)

    def test_init_invalid_ddr(self):
        ''' Initialize with invalid ddr. '''
        with self.assertRaisesRegexp(TypeError, 'VoltageDomain: .*ddr.*'):
            energydram.VoltageDomain(self.tck, self.vdd, self.idds,
                                     self.chipcnt, 0.5)

        with self.assertRaisesRegexp(ValueError, 'VoltageDomain: .*ddr.*'):
            energydram.VoltageDomain(self.tck, self.vdd, self.idds,
                                     self.chipcnt, -1)

    def test_background_energy(self):
        ''' Calculate background energy. '''
        vdom = energydram.VoltageDomain(self.tck, self.vdd, self.idds,
                                        self.chipcnt, ddr=3)

        pds_pre_lo = vdom.background_energy(1, 0, 0, 0) / vdom.tck
        self.assertAlmostEqual(pds_pre_lo, 55.1, delta=0.1)

        pds_pre_hi = vdom.background_energy(0, 1, 0, 0) / vdom.tck
        self.assertAlmostEqual(pds_pre_hi, 66.2, delta=0.1)

        pds_act_lo = vdom.background_energy(0, 0, 1, 0) / vdom.tck
        self.assertAlmostEqual(pds_act_lo, 63.0, delta=0.1)

        pds_act_hi = vdom.background_energy(0, 0, 0, 1) / vdom.tck
        self.assertAlmostEqual(pds_act_hi, 70.9, delta=0.1)

    def test_activate_energy(self):
        ''' Calculate activate energy. '''
        vdom = energydram.VoltageDomain(self.tck, self.vdd, self.idds,
                                        self.chipcnt, ddr=3)

        eact = vdom.activate_energy(self.timing, 1)
        pds_act = eact / (self.timing.RAS + self.timing.RP) / vdom.tck
        self.assertAlmostEqual(pds_act, 80.0, delta=0.1)

    def test_readwrite_energy(self):
        ''' Calculate read/write energy. '''
        vdom = energydram.VoltageDomain(self.tck, self.vdd, self.idds,
                                        self.chipcnt, ddr=3)

        erd = vdom.readwrite_energy(num_rd=1, num_wr=0)
        pds_rd = erd / 4 / vdom.tck
        self.assertAlmostEqual(pds_rd, 212.6, delta=0.1)

        ewr = vdom.readwrite_energy(num_rd=0, num_wr=1)
        pds_wr = ewr / 4 / vdom.tck
        self.assertAlmostEqual(pds_wr, 220.5, delta=0.1)

    def test_refresh_energy(self):
        ''' Calculate activate energy. '''
        vdom = energydram.VoltageDomain(self.tck, self.vdd, self.idds,
                                        self.chipcnt, ddr=3)

        eref = vdom.refresh_energy(self.timing, 1)
        pds_ref = eref / self.timing.REFI / vdom.tck
        self.assertAlmostEqual(pds_ref, 5.5, delta=0.1)




""" $lic$
Copyright (c) 2016-2017, Mingyu Gao
All rights reserved.

This program is free software: you can redistribute it and/or modify it under
the terms of the Modified BSD-3 License as published by the Open Source
Initiative.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the BSD-3 License for more details.

You should have received a copy of the Modified BSD-3 License along with this
program. If not, see <https://opensource.org/licenses/BSD-3-Clause>.
"""

import os
import re
import setuptools

PACKAGE = 'energydram'
DESC = 'DRAM energy calculation.'

def _get_version():
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, PACKAGE, '__init__.py'), 'r') as fh:
        matches = re.findall(r'^\s*__version__\s*=\s*[\'"]([^\'"]+)[\'"]',
                             fh.read(), re.M)
        if matches:
            return matches[-1]
    return '0.0.0'

def _readme():
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, 'README.rst'), 'r') as fh:
        return fh.read()

setuptools.setup(
    name=PACKAGE,
    version=_get_version(),
    description=DESC,

    author='Mingyu Gao',
    author_email='mgao12@stanford.edu',
    long_description=_readme(),
    url='https://github.com/gaomy3832/energydram',
    license='BSD 3-clause',

    packages=setuptools.find_packages(),

    install_requires=[
        'numpy>=1.8',
        'coverage>=4',
        'pytest>=3',
        'pytest-cov>=2',
        'pytest-xdist>=1',
    ],

    keywords='DRAM energy',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: System :: Hardware',
        'Topic :: Utilities',
    ],
)




""" $lic$
Copyright (c) 2016-2017, Mingyu Gao
All rights reserved.

This program is free software: you can redistribute it and/or modify it under
the terms of the Modified BSD-3 License as published by the Open Source
Initiative.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the BSD-3 License for more details.

You should have received a copy of the Modified BSD-3 License along with this
program. If not, see <https://opensource.org/licenses/BSD-3-Clause>.
"""

import unittest

import energydram


class TestEnergyDDR(unittest.TestCase):
    '''
    EnergyDDR class unit tests.

    Based on LPDDR3, 8 Gb, x32, dual-die, speed 1600, fast-exit
    (178b_8-16gb_2c0f_mobile_lpddr3.pdf).
    '''

    tck = 1000./800
    timing = energydram.Timing(RRD=10, RAS=42, RP=18, RFC=210, REFI=3900)
    vdd1 = 1.8
    idds1 = energydram.IDDs(idd0=8, idd2p=0.8, idd2n=0.8, idd3p=1.4,
                            idd3n=2.0, idd4r=2, idd4w=2, idd5=28)
    vdd2 = 1.2
    idds2 = energydram.IDDs(idd0=60, idd2p=1.8, idd2n=26, idd3p=11,
                            idd3n=34, idd4r=230, idd4w=240, idd5=150)
    vddcaq = 1.2
    # NOTE: idd0, idd4r, idd4w, idd5 were 3.
    iddsin = energydram.IDDs(idd0=6, idd2p=0.2, idd2n=6, idd3p=0.2,
                             idd3n=6, idd4r=6, idd4w=6, idd5=6)
    chipcnt = 1

    def test_init(self):
        ''' Initialization. '''
        elpddr3 = energydram.EnergyLPDDR(self.tck, self.timing, self.vdd1,
                                         self.idds1, self.vdd2, self.idds2,
                                         self.vddcaq, self.iddsin, self.chipcnt,
                                         ddr=3)
        self.assertEqual(elpddr3.vdom1.tck, self.tck, 'tck1')
        self.assertEqual(elpddr3.vdom2.tck, self.tck, 'tck2')
        self.assertEqual(elpddr3.vdomcaq.tck, self.tck, 'tckcaq')
        self.assertEqual(elpddr3.timing, self.timing, 'timing')
        self.assertEqual(elpddr3.vdom1.vdd, self.vdd1, 'vdd1')
        self.assertEqual(elpddr3.vdom2.vdd, self.vdd2, 'vdd2')
        self.assertEqual(elpddr3.vdomcaq.vdd, self.vddcaq, 'vddcaq')
        self.assertEqual(elpddr3.vdom1.idds, self.idds1, 'idds1')
        self.assertEqual(elpddr3.vdom2.idds, self.idds2, 'idds2')
        self.assertEqual(elpddr3.vdomcaq.idds, self.iddsin, 'iddsin')
        self.assertEqual(elpddr3.vdom1.chipcnt, self.chipcnt, 'chipcnt1')
        self.assertEqual(elpddr3.vdom2.chipcnt, self.chipcnt, 'chipcnt2')
        self.assertEqual(elpddr3.vdomcaq.chipcnt, self.chipcnt, 'chipcntcaq')
        self.assertEqual(elpddr3.vdom1.ddr, 3, 'ddr1')
        self.assertEqual(elpddr3.vdom2.ddr, 3, 'ddr2')
        self.assertEqual(elpddr3.vdomcaq.ddr, 3, 'ddrcaq')
        self.assertIn('LPDDR3', elpddr3.type, 'type')

    def test_init_lpddr2(self):
        ''' Initialization for LPDDR2. '''
        elpddr2 = energydram.EnergyLPDDR(self.tck, self.timing, self.vdd1,
                                         self.idds1, self.vdd2, self.idds2,
                                         self.vddcaq, self.iddsin, self.chipcnt,
                                         ddr=2)
        self.assertIn('LPDDR2', elpddr2.type, 'type')

    def test_init_lpddr(self):
        ''' Initialize with for LPDDR. '''
        with self.assertRaisesRegexp(ValueError, 'EnergyLPDDR: .*ddr.*'):
            energydram.EnergyLPDDR(self.tck, self.timing, self.vdd1,
                                   self.idds1, self.vdd2, self.idds2,
                                   self.vddcaq, self.iddsin, self.chipcnt,
                                   ddr=1)

    def test_init_lpddr4(self):
        ''' Initialize with for LPDDR4. '''
        with self.assertRaisesRegexp(ValueError, 'EnergyLPDDR: .*ddr.*'):
            energydram.EnergyLPDDR(self.tck, self.timing, self.vdd1,
                                   self.idds1, self.vdd2, self.idds2,
                                   self.vddcaq, self.iddsin, self.chipcnt,
                                   ddr=4)

    def test_init_lpddr23_invalid_vdd1(self):
        ''' Initialize with invalid vdd1 for LPDDR2/3. '''
        with self.assertRaisesRegexp(ValueError, 'EnergyLPDDR: .*vdd1.*'):
            energydram.EnergyLPDDR(self.tck, self.timing, 1.2,
                                   self.idds1, self.vdd2, self.idds2,
                                   self.vddcaq, self.iddsin, self.chipcnt)

    def test_init_lpddr23_invalid_vdd2(self):
        ''' Initialize with invalid vdd2 for LPDDR2/3. '''
        with self.assertRaisesRegexp(ValueError, 'EnergyLPDDR: .*vdd2.*'):
            energydram.EnergyLPDDR(self.tck, self.timing, self.vdd1,
                                   self.idds1, 1.8, self.idds2,
                                   self.vddcaq, self.iddsin, self.chipcnt)

    def test_init_lpddr23_invalid_vddcaq(self):  # pylint: disable=invalid-name
        ''' Initialize with invalid vddcaq for LPDDR2/3. '''
        with self.assertRaisesRegexp(ValueError, 'EnergyLPDDR: .*vddcaq.*'):
            energydram.EnergyLPDDR(self.tck, self.timing, self.vdd1,
                                   self.idds1, self.vdd2, self.idds2,
                                   1.8, self.iddsin, self.chipcnt)

    def test_init_lpddr23_invalid_timing(self):  # pylint: disable=invalid-name
        ''' Initialize with invalid timing for LPDDR2/3. '''
        with self.assertRaisesRegexp(TypeError, 'EnergyLPDDR: .*timing.*'):
            energydram.EnergyLPDDR(self.tck, None, self.vdd1,
                                   self.idds1, self.vdd2, self.idds2,
                                   self.vddcaq, self.iddsin, self.chipcnt)

    def test_background_energy(self):
        ''' Calculate background energy. '''
        elpddr3 = energydram.EnergyLPDDR(self.tck, self.timing, self.vdd1,
                                         self.idds1, self.vdd2, self.idds2,
                                         self.vddcaq, self.iddsin, self.chipcnt,
                                         ddr=3)

        pds_pre_lo = elpddr3.background_energy(1, 0, 0, 0) / self.tck
        self.assertIsInstance(pds_pre_lo, float)

        pds_pre_hi = elpddr3.background_energy(0, 1, 0, 0) / self.tck
        self.assertIsInstance(pds_pre_hi, float)

        pds_act_lo = elpddr3.background_energy(0, 0, 1, 0) / self.tck
        self.assertIsInstance(pds_act_lo, float)

        pds_act_hi = elpddr3.background_energy(0, 0, 0, 1) / self.tck
        self.assertIsInstance(pds_act_hi, float)

    def test_activate_energy(self):
        ''' Calculate activate energy. '''
        elpddr3 = energydram.EnergyLPDDR(self.tck, self.timing, self.vdd1,
                                         self.idds1, self.vdd2, self.idds2,
                                         self.vddcaq, self.iddsin, self.chipcnt,
                                         ddr=3)

        eact = elpddr3.activate_energy(1)
        pds_act = eact / (self.timing.RAS + self.timing.RP) / self.tck
        self.assertIsInstance(pds_act, float)

    def test_readwrite_energy(self):
        ''' Calculate read/write energy. '''
        elpddr3 = energydram.EnergyLPDDR(self.tck, self.timing, self.vdd1,
                                         self.idds1, self.vdd2, self.idds2,
                                         self.vddcaq, self.iddsin, self.chipcnt,
                                         ddr=3)

        erd = elpddr3.readwrite_energy(num_rd=1, num_wr=0)
        pds_rd = erd / 4 / self.tck
        self.assertIsInstance(pds_rd, float)

        ewr = elpddr3.readwrite_energy(num_rd=0, num_wr=1)
        pds_wr = ewr / 4 / self.tck
        self.assertIsInstance(pds_wr, float)

    def test_refresh_energy(self):
        ''' Calculate activate energy. '''
        elpddr3 = energydram.EnergyLPDDR(self.tck, self.timing, self.vdd1,
                                         self.idds1, self.vdd2, self.idds2,
                                         self.vddcaq, self.iddsin, self.chipcnt,
                                         ddr=3)

        eref = elpddr3.refresh_energy(1)
        pds_ref = eref / self.timing.REFI / self.tck
        self.assertIsInstance(pds_ref, float)




""" $lic$
Copyright (c) 2016-2017, Mingyu Gao
All rights reserved.

This program is free software: you can redistribute it and/or modify it under
the terms of the Modified BSD-3 License as published by the Open Source
Initiative.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the BSD-3 License for more details.

You should have received a copy of the Modified BSD-3 License along with this
program. If not, see <https://opensource.org/licenses/BSD-3-Clause>.
"""

import unittest

import energydram


class TestTermResistance(unittest.TestCase):
    ''' TermResistance class unit tests. '''

    def test_init(self):
        ''' Initialization. '''
        resistance = energydram.TermResistance(rz_dev=34, rz_mc=34, rtt_nom=30,
                                               rtt_wr=120, rtt_mc=75, rs=15)
        self.assertEqual(resistance.rz_dev, 34, 'R_z')
        self.assertEqual(resistance.rz_mc, 34, 'R_zC')
        self.assertEqual(resistance.rtt_nom, 30, 'R_TT,nom')
        self.assertEqual(resistance.rtt_wr, 120, 'R_TT(WR)')
        self.assertEqual(resistance.rtt_mc, 75, 'R_TTC')
        self.assertEqual(resistance.rs, 15, 'R_s')

    def test_init_zero_rz_dev(self):
        ''' Assert rz_dev != 0. '''
        with self.assertRaisesRegexp(ValueError, 'TermResistance: .*rz_dev.*'):
            energydram.TermResistance(rz_dev=0, rz_mc=34, rtt_nom=30,
                                      rtt_wr=120, rtt_mc=75, rs=15)

    def test_init_negative_rz_dev(self):
        ''' Assert rz_dev > 0. '''
        with self.assertRaisesRegexp(ValueError, 'TermResistance: .*rz_dev.*'):
            energydram.TermResistance(rz_dev=-10, rz_mc=34, rtt_nom=30,
                                      rtt_wr=120, rtt_mc=75, rs=15)

    def test_init_zero_rz_mc(self):
        ''' Assert rz_mc != 0. '''
        with self.assertRaisesRegexp(ValueError, 'TermResistance: .*rz_mc.*'):
            energydram.TermResistance(rz_dev=34, rz_mc=0, rtt_nom=30,
                                      rtt_wr=120, rtt_mc=75, rs=15)

    def test_init_negative_rz_mc(self):
        ''' Assert rz_mc > 0. '''
        with self.assertRaisesRegexp(ValueError, 'TermResistance: .*rz_mc.*'):
            energydram.TermResistance(rz_dev=34, rz_mc=-10, rtt_nom=30,
                                      rtt_wr=120, rtt_mc=75, rs=15)

    def test_init_zero_rtt_nom(self):
        ''' Assert rtt_nom != 0. '''
        with self.assertRaisesRegexp(ValueError,
                                     'TermResistance: .*R_TT,nom.*'):
            energydram.TermResistance(rz_dev=34, rz_mc=34, rtt_nom=0,
                                      rtt_wr=120, rtt_mc=75, rs=15)

    def test_init_negative_rtt_nom(self):
        ''' Assert rtt_nom > 0. '''
        with self.assertRaisesRegexp(ValueError,
                                     'TermResistance: .*R_TT,nom.*'):
            energydram.TermResistance(rz_dev=34, rz_mc=34, rtt_nom=-10,
                                      rtt_wr=120, rtt_mc=75, rs=15)

    def test_init_zero_rtt_wr(self):
        ''' Assert rtt_wr != 0. '''
        with self.assertRaisesRegexp(ValueError,
                                     r'TermResistance: .*R_TT\(WR\).*'):
            energydram.TermResistance(rz_dev=34, rz_mc=34, rtt_nom=30,
                                      rtt_wr=0, rtt_mc=75, rs=15)

    def test_init_negative_rtt_wr(self):
        ''' Assert rtt_wr > 0. '''
        with self.assertRaisesRegexp(ValueError,
                                     r'TermResistance: .*R_TT\(WR\).*'):
            energydram.TermResistance(rz_dev=34, rz_mc=34, rtt_nom=30,
                                      rtt_wr=-10, rtt_mc=75, rs=15)

    def test_init_zero_rtt_mc(self):
        ''' Assert rtt_mc != 0. '''
        with self.assertRaisesRegexp(ValueError, 'TermResistance: .*rtt_mc.*'):
            energydram.TermResistance(rz_dev=34, rz_mc=34, rtt_nom=30,
                                      rtt_wr=120, rtt_mc=0, rs=15)

    def test_init_negative_rtt_mc(self):
        ''' Assert rtt_mc > 0. '''
        with self.assertRaisesRegexp(ValueError, 'TermResistance: .*rtt_mc.*'):
            energydram.TermResistance(rz_dev=34, rz_mc=34, rtt_nom=30,
                                      rtt_wr=120, rtt_mc=-10, rs=15)

    def test_init_zero_rs(self):
        ''' Assert rs != 0. '''
        with self.assertRaisesRegexp(ValueError, 'TermResistance: .*rs.*'):
            energydram.TermResistance(rz_dev=34, rz_mc=34, rtt_nom=30,
                                      rtt_wr=120, rtt_mc=75, rs=0)

    def test_init_negative_rs(self):
        ''' Assert rs > 0. '''
        with self.assertRaisesRegexp(ValueError, 'TermResistance: .*rs.*'):
            energydram.TermResistance(rz_dev=34, rz_mc=34, rtt_nom=30,
                                      rtt_wr=120, rtt_mc=75, rs=-10)


class TestTermination(unittest.TestCase):
    '''
    Termination class unit tests.

    Compare with DDR3_Power_Calc.xlsm, default setting but Rs1 = 15 for write.
    '''

    vdd = 1.5
    rankcnt = 2
    resistance = energydram.TermResistance(rz_dev=34, rz_mc=34, rtt_nom=40,
                                           rtt_wr=30, rtt_mc=60, rs=15)

    def test_init(self):
        ''' Initialization. '''
        term = energydram.Termination(self.vdd, self.rankcnt, self.resistance)
        self.assertEqual(term.vdd, self.vdd, 'vdd')
        self.assertEqual(term.rankcnt, self.rankcnt, 'rankcnt')
        self.assertEqual(term.resistance, self.resistance, 'resistance')
        self.assertEqual(term.rdpincnt, 1, 'rdpincnt')
        self.assertEqual(term.wrpincnt, 1, 'wrpincnt')

    def test_init_width(self):
        ''' Initialize with different width values. '''
        term = energydram.Termination(self.vdd, self.rankcnt, self.resistance,
                                      width=4)
        self.assertEqual(term.rdpincnt, 6, 'rdpincnt')
        self.assertEqual(term.wrpincnt, 7, 'wrpincnt')
        term = energydram.Termination(self.vdd, self.rankcnt, self.resistance,
                                      width=8)
        self.assertEqual(term.rdpincnt, 10, 'rdpincnt')
        self.assertEqual(term.wrpincnt, 11, 'wrpincnt')
        term = energydram.Termination(self.vdd, self.rankcnt, self.resistance,
                                      width=16)
        self.assertEqual(term.rdpincnt, 20, 'rdpincnt')
        self.assertEqual(term.wrpincnt, 22, 'wrpincnt')

    def test_init_invalid_vdd(self):
        ''' Initialize with invalid vdd. '''
        with self.assertRaisesRegexp(ValueError, 'Termination: .*vdd.*'):
            energydram.Termination(-1.2, self.rankcnt, self.resistance)

    def test_init_invalid_rankcnt(self):
        ''' Initialize with invalid rankcnt. '''
        with self.assertRaisesRegexp(TypeError, 'Termination: .*rankcnt.*'):
            energydram.Termination(self.vdd, 1.2, self.resistance)

        with self.assertRaisesRegexp(ValueError, 'Termination: .*rankcnt.*'):
            energydram.Termination(self.vdd, 0, self.resistance)

    def test_init_invalid_width(self):
        ''' Initialize with invalid width. '''
        with self.assertRaisesRegexp(ValueError, 'Termination: .*width.*'):
            energydram.Termination(self.vdd, self.rankcnt, self.resistance,
                                   width=1)

    def test_init_invalid_resistance(self):
        ''' Initialize with invalid resistance. '''
        with self.assertRaisesRegexp(TypeError, 'Termination: .*resistance.*'):
            energydram.Termination(self.vdd, self.rankcnt, None)

    def test_init_invalid_level(self):
        ''' Initialize with invalid level. '''
        with self.assertRaisesRegexp(ValueError, 'Termination: .*level.*'):
            energydram.Termination(self.vdd, self.rankcnt, self.resistance,
                                   level='inv')

    def test_power_total(self):
        ''' Assert total power is the sum of mem controller and devices. '''
        term = energydram.Termination(self.vdd, self.rankcnt, self.resistance)
        self.assertAlmostEqual(
            term.read_power_total(),
            term.read_power_memctlr() + term.read_power_devices(),
            places=6)
        self.assertAlmostEqual(
            term.write_power_total(),
            term.write_power_memctlr() + term.write_power_devices(),
            places=6)

    def test_power_devices(self):
        ''' Assert DRAM device power is the sum of all ranks. '''
        term = energydram.Termination(self.vdd, self.rankcnt, self.resistance)
        self.assertAlmostEqual(
            term.read_power_devices(),
            term.read_power_target_rank() + term.read_power_other_ranks(),
            places=6)
        self.assertAlmostEqual(
            term.write_power_devices(),
            term.write_power_target_rank() + term.write_power_other_ranks(),
            places=6)

    def test_power_memctlr(self):
        ''' Calculate memory controller power. '''
        term = energydram.Termination(self.vdd, self.rankcnt, self.resistance)
        self.assertAlmostEqual(term.read_power_memctlr(), 10.7e-3,
                               places=4)
        self.assertAlmostEqual(term.write_power_memctlr(), 5.5e-3,
                               places=4)

    def test_power_target_rank(self):
        ''' Calculate DRAM device power at the target rank. '''
        term = energydram.Termination(self.vdd, self.rankcnt, self.resistance)
        self.assertAlmostEqual(term.read_power_target_rank(), 3.2e-3 + 1.4e-3,
                               places=4)
        self.assertAlmostEqual(term.write_power_target_rank(), 20.2e-3 + .74e-3,
                               places=4)

    def test_power_other_ranks(self):
        ''' Calculate DRAM device power at other ranks. '''
        term = energydram.Termination(self.vdd, self.rankcnt, self.resistance)
        self.assertAlmostEqual(term.read_power_other_ranks(), 15e-3 + 0.48e-3,
                               places=4)
        self.assertAlmostEqual(term.write_power_other_ranks(), 15.4e-3 + 0.5e-3,
                               places=4)


class TestTerminationDDR4(unittest.TestCase):
    '''
    Termination class unit tests for DDR4.

    Compare with DDR4_Power_Calc.xlsm, default setting but RTTu2 = 40 for write.
    '''

    vdd = 1.2
    rankcnt = 2
    resistance = energydram.TermResistance(rz_dev=34, rz_mc=34, rtt_nom=40,
                                           rtt_wr=120, rtt_mc=120, rs=10)
    term = energydram.Termination(vdd, rankcnt, resistance,
                                  level='high')

    def test_power_memctlr(self):
        ''' Calculate memory controller power. '''
        self.assertAlmostEqual(self.term.read_power_memctlr(), 2.4e-3,
                               places=4)
        self.assertAlmostEqual(self.term.write_power_memctlr(), 10.0e-3,
                               places=4)

    def test_power_target_rank(self):
        ''' Calculate DRAM device power at the target rank. '''
        self.assertAlmostEqual(self.term.read_power_target_rank(),
                               7.8e-3 + 2.3e-3,
                               places=4)
        self.assertAlmostEqual(self.term.write_power_target_rank(),
                               2.7e-3 + 0.2e-3,
                               places=4)

    def test_power_other_ranks(self):
        ''' Calculate DRAM device power at other ranks. '''
        self.assertAlmostEqual(self.term.read_power_other_ranks(),
                               4.6e-3 + 1.1e-3,
                               places=4)
        self.assertAlmostEqual(self.term.write_power_other_ranks(),
                               6.1e-3 + 1.5e-3,
                               places=4)


class TestTerminationLPDDR3(unittest.TestCase):
    '''
    Termination class unit tests for LPDDR3.

    Compare with Kenta & Makoto Excel sheet.
    '''

    vdd = 1.2
    rankcnt = 2
    # Ignore 5 pF on the P2P wire, but keep the 50 ohm.
    resistance = energydram.TermResistance(rz_dev=40, rz_mc=40, rtt_nom=60,
                                           rtt_wr=60, rtt_mc=60, rs=50)
    term = energydram.Termination(vdd, rankcnt, resistance, width=16,
                                  level='high')

    def test_power_target_rank(self):
        ''' Calculate DRAM device power at the target rank. '''
        self.assertAlmostEqual(self.term.read_power_target_rank(),
                               156.2e-3,
                               places=4)
        self.assertAlmostEqual(self.term.write_power_target_rank(),
                               96.5e-3,
                               places=4)

    def test_power_other_ranks(self):
        ''' Calculate DRAM device power at other ranks. '''
        self.assertAlmostEqual(self.term.read_power_other_ranks(),
                               23.8e-3,
                               places=4)
        self.assertAlmostEqual(self.term.write_power_other_ranks(),
                               96.5e-3,
                               places=4)


class TestTerminationGDDR5(unittest.TestCase):
    '''
    Termination class unit tests for GDDR5.

    Compare with Kenta & Makoto Excel sheet.
    '''

    vdd = 1.5
    rankcnt = 2
    resistance = energydram.TermResistance(rz_dev=40, rz_mc=40, rtt_nom=60,
                                           rtt_wr=60, rtt_mc=60, rs=15)
    term = energydram.Termination(vdd, rankcnt, resistance, width=32,
                                  with_dqs=False, with_dm=False, with_dbi=True,
                                  level='high')

    def test_power_target_rank(self):
        ''' Calculate DRAM device power at the target rank. '''
        self.assertAlmostEqual(self.term.read_power_target_rank(),
                               317.2e-3,
                               places=4)
        self.assertAlmostEqual(self.term.write_power_target_rank(),
                               140.5e-3,
                               places=4)

    def test_power_other_ranks(self):
        ''' Calculate DRAM device power at other ranks. '''
        self.assertAlmostEqual(self.term.read_power_other_ranks(),
                               85.4e-3,
                               places=4)
        self.assertAlmostEqual(self.term.write_power_other_ranks(),
                               140.5e-3,
                               places=4)




""" $lic$
Copyright (c) 2016-2017, Mingyu Gao
All rights reserved.

This program is free software: you can redistribute it and/or modify it under
the terms of the Modified BSD-3 License as published by the Open Source
Initiative.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the BSD-3 License for more details.

You should have received a copy of the Modified BSD-3 License along with this
program. If not, see <https://opensource.org/licenses/BSD-3-Clause>.
"""

import unittest

import energydram


class TestEnergyDDR(unittest.TestCase):
    '''
    EnergyDDR class unit tests.

    Compare with DDR3_Power_Calc.xlsm.

    Based on DDR3, 2 Gb, x8, -125E, fast-exit.
    '''

    tck = 1000./800
    timing = energydram.Timing(RRD=6, RAS=35, RP=47.5-35, RFC=160, REFI=7800)
    vdd = 1.5
    idds = energydram.IDDs(idd0=95, idd2p=35, idd2n=42, idd3p=40,
                           idd3n=45, idd4r=180, idd4w=185, idd5=215)
    chipcnt = 1

    def test_init(self):
        ''' Initialization. '''
        eddr3 = energydram.EnergyDDR(self.tck, self.timing, self.vdd,
                                     self.idds, self.chipcnt)
        self.assertEqual(eddr3.vdom.tck, self.tck, 'tck')
        self.assertEqual(eddr3.timing, self.timing, 'timing')
        self.assertEqual(eddr3.vdom.vdd, self.vdd, 'vdd')
        self.assertEqual(eddr3.vdom.idds, self.idds, 'idds')
        self.assertEqual(eddr3.vdom.chipcnt, self.chipcnt, 'chipcnt')
        self.assertEqual(eddr3.vdom.ddr, 3, 'ddr')
        self.assertIn('DDR3', eddr3.type, 'type')

    def test_init_ddr2(self):
        ''' Initialization for DDR2. '''
        eddr2 = energydram.EnergyDDR(self.tck, self.timing, 1.8,
                                     self.idds, self.chipcnt, ddr=2)
        self.assertIn('DDR2', eddr2.type, 'type')

    def test_init_ddr3l(self):
        ''' Initialization for DDR3L. '''
        eddr3 = energydram.EnergyDDR(self.tck, self.timing, 1.35,
                                     self.idds, self.chipcnt)
        self.assertIn('DDR3L', eddr3.type, 'type')

    def test_init_ddr(self):
        ''' Initialize with for DDR. '''
        with self.assertRaisesRegexp(ValueError, 'EnergyDDR: .*ddr.*'):
            energydram.EnergyDDR(self.tck, self.timing, 1.2, self.idds,
                                 self.chipcnt, ddr=1)

    def test_init_ddr4(self):
        ''' Initialize with for DDR4. '''
        with self.assertRaisesRegexp(ValueError, 'EnergyDDR: .*ddr.*'):
            energydram.EnergyDDR(self.tck, self.timing, 1.2, self.idds,
                                 self.chipcnt, ddr=4)

    def test_init_ddr2_invalid_vdd(self):
        ''' Initialize with invalid vdd for DDR2. '''
        with self.assertRaisesRegexp(ValueError, 'EnergyDDR: .*vdd.*'):
            energydram.EnergyDDR(self.tck, self.timing, 1.2, self.idds,
                                 self.chipcnt, ddr=2)

    def test_init_ddr3_invalid_vdd(self):
        ''' Initialize with invalid vdd for DDR3. '''
        with self.assertRaisesRegexp(ValueError, 'EnergyDDR: .*vdd.*'):
            energydram.EnergyDDR(self.tck, self.timing, 1.2, self.idds,
                                 self.chipcnt)

    def test_init_invalid_timing(self):
        ''' Initialize with invalid idds. '''
        with self.assertRaisesRegexp(TypeError, 'EnergyDDR: .*timing.*'):
            energydram.EnergyDDR(self.tck, (1, 2), self.vdd, self.idds,
                                 self.chipcnt)

    def test_background_energy(self):
        ''' Calculate background energy. '''
        eddr3 = energydram.EnergyDDR(self.tck, self.timing, self.vdd,
                                     self.idds, self.chipcnt)

        pds_pre_lo = eddr3.background_energy(1, 0, 0, 0) / self.tck
        self.assertAlmostEqual(pds_pre_lo, 52.5, delta=0.1)

        pds_pre_hi = eddr3.background_energy(0, 1, 0, 0) / self.tck
        self.assertAlmostEqual(pds_pre_hi, 63.0, delta=0.1)

        pds_act_lo = eddr3.background_energy(0, 0, 1, 0) / self.tck
        self.assertAlmostEqual(pds_act_lo, 60.0, delta=0.1)

        pds_act_hi = eddr3.background_energy(0, 0, 0, 1) / self.tck
        self.assertAlmostEqual(pds_act_hi, 67.5, delta=0.1)

    def test_activate_energy(self):
        ''' Calculate activate energy. '''
        eddr3 = energydram.EnergyDDR(self.tck, self.timing, self.vdd,
                                     self.idds, self.chipcnt)

        eact = eddr3.activate_energy(1)
        pds_act = eact / (self.timing.RAS + self.timing.RP) / self.tck
        self.assertAlmostEqual(pds_act, 76.2, delta=0.1)

    def test_readwrite_energy(self):
        ''' Calculate read/write energy. '''
        eddr3 = energydram.EnergyDDR(self.tck, self.timing, self.vdd,
                                     self.idds, self.chipcnt)

        erd = eddr3.readwrite_energy(num_rd=1, num_wr=0)
        pds_rd = erd / 4 / self.tck
        self.assertAlmostEqual(pds_rd, 202.5, delta=0.1)

        ewr = eddr3.readwrite_energy(num_rd=0, num_wr=1)
        pds_wr = ewr / 4 / self.tck
        self.assertAlmostEqual(pds_wr, 210.0, delta=0.1)

    def test_refresh_energy(self):
        ''' Calculate activate energy. '''
        eddr3 = energydram.EnergyDDR(self.tck, self.timing, self.vdd,
                                     self.idds, self.chipcnt)

        eref = eddr3.refresh_energy(1)
        pds_ref = eref / self.timing.REFI / self.tck
        self.assertAlmostEqual(pds_ref, 5.2, delta=0.1)

